#include "globals.c"
#include "macros.c"
#include "structs.c"

#define DISK_RETRY_COUNT 16

static void save_rawcache(void);

extern unsigned char __far *buffer_segment;

static void lba2chs(unsigned long index, struct disk *disk, struct chs *retval)
{
//  retval->sector = 1 + (index % disk->sectors);
//  retval->head = (index / disk->sectors) % disk->heads;
//  retval->cylinder = index / (disk->sectors * disk->heads);
  retval->sector = 1 + (LONGMOD(index, disk->sectors));
  retval->head = LONGDIV(index, disk->sectors);
  retval->head = LONGMOD(retval->head, disk->heads);
  retval->cylinder = LONGDIV(index, disk->sectors * disk->heads);
}


unsigned read_sector_pragma(unsigned cylinder_sector, unsigned head_drive, unsigned count, void __far *buff);
#pragma aux read_sector_pragma = \
  "xchg ch, cl" \
  "ror cl, 1" \
  "ror cl, 1" \
  "mov ah, 0x02" \
  "int 0x13" \
  parm [cx] [dx] [ax] [es bx] value [ax];
// (dx must be saved for bug compatibility with some BIOS implementations)

unsigned write_sector_pragma(unsigned cylinder_sector, unsigned head_drive, unsigned count, void __far *buff);
#pragma aux write_sector_pragma = \
  "xchg ch, cl" \
  "ror cl, 1" \
  "ror cl, 1" \
  "mov ah, 0x03" \
  "int 0x13" \
  parm [cx] [dx] [ax] [es bx] value [ax];

unsigned verify_sector_pragma(unsigned cylinder_sector, unsigned head_drive, unsigned count, void __far *buff);
#pragma aux verify_sector_pragma = \
  "xchg ch, cl" \
  "ror cl, 1" \
  "ror cl, 1" \
  "mov ah, 0x04" \
  "int 0x13" \
  parm [cx] [dx] [ax] [es bx] value [ax];

unsigned reset_disk_pragma(unsigned char disk_id);
#pragma aux reset_disk_pragma = \
  "xor ah, ah" \
  "int 0x13" \
  parm [dl] value [ax] modify exact [];  

  
unsigned long get_clock_ticks(void);
#pragma aux get_clock_ticks = \
  "push ax" \
  "xor ah, ah" \
  "int 0x1A" \
  "pop ax" \
  value [cx dx];

extern unsigned long diskette_last_access[]; 

int disk_changed(unsigned char disk_id, unsigned long *clockptr)
{
  unsigned long clock;

  if(disk_id & 0x80) return 0; // assume hard disk doesn't change

  clock = get_clock_ticks();
  if(clockptr) *clockptr = clock;

  if(!WRITE_FLAG && (clock - diskette_last_access[disk_id] > 2*TICKS_PER_SECOND))
    return 1;
  return 0;
}

static void invalidate_rawcache();

#include "caches.c"

static void set_rawcache_last(lbaindex_t sector, unsigned char disk_id, unsigned char count)
{
  RAWCACHE_LAST_DISK = disk_id;
  RAWCACHE_LAST_SECTOR = sector;
  RAWCACHE_LAST_SECTOR_COUNT = count;
}

inline void change_dpt(unsigned char sectors)
{
  ALT_DPT[4] = sectors;
}

static unsigned get_disk_parm(unsigned char disk_id, struct disk *retval, unsigned nochgchk)
{
  unsigned status;
  unsigned long clock;
  unsigned char *buff = rawcache_buff;
  
  if(!(disk_id & 0x80))
  {
    // diskette changed?
    if((disk_changed(disk_id, &clock) && !nochgchk) || !DRIVE_SECTORS(disk_id))
    {
      if(cache_seg) invalidate_caches(disk_id);
      status = reset_disk_pragma(disk_id) >> 8;
      if(status) return status;
      status = read_sector_pragma(1 << 10, disk_id, 1, buff) >> 8; // read boot sector to buffer
      if(status)
      {
        return status;
      }
      set_rawcache_last(0, disk_id, 1);
      DRIVE_SECTORS(disk_id) = *(unsigned __far*)&buff[24];
      DRIVE_HEADS(disk_id) = *(unsigned __far*)&buff[26];
    }
    diskette_last_access[disk_id] = clock;
    change_dpt(retval->sectors = DRIVE_SECTORS(disk_id));
    retval->heads = DRIVE_HEADS(disk_id);
    return 0;
  }

  disk_id &= 0x7F;
  disk_id += FLOPPY_COUNT_R;
  retval->sectors = DRIVE_SECTORS(disk_id);
  retval->heads = DRIVE_HEADS(disk_id);
 
  return 0;
}
  
static unsigned read_physical_sector(unsigned char disk_id, struct chs *chs, unsigned count)
{
  unsigned cylinder = chs->cylinder;
  unsigned sector = chs->sector;
  unsigned head = chs->head;

  return read_sector_pragma(cylinder | (sector << 10), (head << 8) | disk_id, count, rawcache_buff);
}

static void create_lba_packet(lbaindex_t index, unsigned char count,
                              unsigned char *packet)
{
  nmemset(packet, 0, 0x10);
  packet[0] = 0x10;
  packet[2] = count;
  *(void __far**)&packet[4] = rawcache_buff;
  *(lbaindex_t*)&packet[8] = index;
}

unsigned read_lba_pragma(unsigned char disk_id, void *packet);
#pragma aux read_lba_pragma = \
  "mov ah, 0x42" \
  "int 0x13" \
  parm [dl] [si] value [ax];

unsigned write_lba_pragma(unsigned char disk_id, void *packet);
#pragma aux write_lba_pragma = \
  "mov ax, 0x4300" \
  "int 0x13" \
  parm [dl] [si] value [ax];

unsigned verify_lba_pragma(unsigned char disk_id, void *packet);
#pragma aux verify_lba_pragma = \
  "mov ah, 0x44" \
  "int 0x13" \
  parm [dl] [si] value [ax];

static unsigned read_lba(lbaindex_t index, unsigned char disk_id,
                         unsigned char count)
{
  unsigned char packet[0x10];
  create_lba_packet(index, count, packet);
  return read_lba_pragma(disk_id, packet) >> 8;
}

static unsigned write_lba(lbaindex_t index, unsigned char disk_id,
                          unsigned char count)
{
  unsigned char packet[0x10];
  unsigned rc;
  create_lba_packet(index, count, packet);
  rc = write_lba_pragma(disk_id, packet) >> 8;
  if(VERIFY_FLAG && !rc)
  {
    create_lba_packet(index, count, packet);
    return verify_lba_pragma(disk_id, packet) >> 8;
  }
  return rc;
}

static unsigned write_physical_sector(unsigned char disk_id, struct chs *chs, unsigned count)
{
  unsigned cylinder = chs->cylinder;
  unsigned sector = chs->sector;
  unsigned head = chs->head;
  unsigned rc;

  rc = write_sector_pragma(cylinder | (sector << 10), (head << 8) | disk_id, count, rawcache_buff);
  
  if(VERIFY_FLAG && !rc)
    return verify_sector_pragma(cylinder | (sector << 10), (head << 8) | disk_id, count, rawcache_buff);

  return rc;
}

static void invalidate_rawcache()
{
  RAWCACHE_LAST_SECTOR_COUNT = 0;
  DYNFLAGS &= ~RAWCACHE_SAVED_BIT;
}

static void save_rawcache()
{
  unsigned char const disk_id = RAWCACHE_LAST_DISK;
  lbaindex_t sector = RAWCACHE_LAST_SECTOR;
  unsigned char count = RAWCACHE_LAST_SECTOR_COUNT;
  unsigned offset = 0;
  unsigned char * const ptr = rawcache_buff;
  unsigned refs = DEFAULT_REFS;
  unsigned const size = DRIVE_SECTOR_SIZE(disk_id);

  if(!count) return;

  if(count > cache_size) count = cache_size;

  while(1)
  {
    save_to_cache(sector, disk_id, count, &ptr[offset], refs | 0x4000);
    if(!--count) break;
    refs++;
    offset += size;
    sector++;
  }

  DYNFLAGS &= ~RAWCACHE_SAVED_BIT;
}

void yield_rawcache(char __far *retstr)
{
  save_rawcache();
  invalidate_rawcache();
  *(unsigned __far*)&retstr[0] = rawcache_size;
  *(void * __far*)&retstr[2] = rawcache_buff;
  *(unsigned __far*)&retstr[4] = get_code_segment();
}

static int disk_critical_error(unsigned char const disk, unsigned writing, unsigned status)
{
  return critical_error_handler(disk, CRITERR_DATAAREA | CRITERR_ALLOW_FAIL
                                | CRITERR_ALLOW_RETRY | CRITERR_ALLOW_IGNORE
                                | CRITERR_ALLOW_ABORT | writing,
                                bios2doserror(status));
}

int write_logical_sector(lbaindex_t sector, unsigned char const disk_id)
{
  struct disk disk;
  struct chs chs;
  unsigned retries;
  unsigned status;
  unsigned sector_size = DRIVE_SECTOR_SIZE(disk_id);
  unsigned char const phys_disk = logical2physicaldrive(disk_id);
  void *rawbuff = rawcache_buff;
  unsigned char __far *buff = buffer_segment;

  retry:
  retries = DISK_RETRY_COUNT;
  do
    status = get_disk_parm(phys_disk, &disk, WRITE_FLAG);
  while(status && --retries);

  if(!retries)
  {
    critical_error:switch(disk_critical_error(disk_id, 1, status))
    {
      case CRITERR_IGNORE: status = 0;
      case CRITERR_FAIL: break;
      case CRITERR_RETRY: goto retry;
    }
    return -status;
  }

  if(VERIFY_FLAG || !cache_seg || (write_to_cache(sector, buff, disk_id) == -1))
  {
    if(cache_seg && !RAWCACHE_SAVED)
      save_rawcache();
    memcpy(rawbuff, buff, sector_size);
    while(1)
    {
      if(disk.sectors != 0xFF)
      {
        lba2chs(sector, &disk, &chs);
        DRIVE_LAST_CYL(disk_id) = chs.cylinder;
        status = write_physical_sector(phys_disk, &chs, 1) >> 8;
      }
      else
        status = write_lba(sector, phys_disk, 1);
      if(!status) break;

      invalidate_rawcache();
      if(!retries--)
        goto critical_error;
      else if(!(retries & 3))
        reset_disk_pragma(phys_disk);      
    }
    if(cache_seg)
    {
      save_to_cache(sector, disk_id, 1, buff, DEFAULT_REFS | 0x8000);
      invalidate_rawcache();
    }
    else
      set_rawcache_last(sector, disk_id, 1);
  }
  else
    if(RAWCACHE_LAST_DISK == disk_id)
      invalidate_rawcache();

  return 0;
}


#define SECTOR_FLUSH_COUNT 18

void flush_caches(const unsigned index)
{
  unsigned n;
  struct cache_entry * const cache_table = DISK_CACHE;
  unsigned count;
  unsigned char disk;
  unsigned sector_table[SECTOR_FLUSH_COUNT];
  unsigned status;
  unsigned char *raw_cache = rawcache_buff;
  struct disk disk_struct;
  struct chs chs;
  unsigned sector_size;
  unsigned char phys_disk;
  unsigned retries;
  unsigned m;
  unsigned max_count;
  unsigned last_cyl;
  unsigned max_cyl_diff;
  unsigned total_count;
  unsigned min_cyl_diff;

  invalidate_rawcache();

  if(index == -1)
  {
    DYNFLAGS &= ~WRITE_BIT;
    disk = FLOPPY_COUNT_R + HARD_DISK_COUNT;
  }
  else
  {
    disk = cache_table[index].drive;
    goto onlylp;
  }

  nextdisk:while(disk--)
  {
    if(!drive_exists(disk) || !(DRIVE_TABLE_ARR[disk*DRIVE_STRUCT_SIZE] & SAVE_CACHE_BIT))
      continue;

    DRIVE_TABLE_ARR[disk*DRIVE_STRUCT_SIZE] &= ~SAVE_CACHE_BIT;
    onlylp:
    phys_disk = logical2physicaldrive(disk);
    get_disk_parm(phys_disk, &disk_struct, 1);

    sector_size = DRIVE_SECTOR_SIZE(disk);
    max_count = rawcache_size / sector_size;
    if(max_count > SECTOR_FLUSH_COUNT)
      max_count = SECTOR_FLUSH_COUNT;

    do
    {
      if(disk_struct.sectors == 0xFF)
        max_cyl_diff = -1;
      else
        max_cyl_diff = 0;
      min_cyl_diff = -1;
      if(index != -1)
      {
        lba2chs(cache_table[index].sector, &disk_struct, &chs);
        last_cyl = chs.cylinder;
      }
      else
        last_cyl = DRIVE_LAST_CYL(disk);
      do
      {
        for(total_count=count=n=0;n<(volatile unsigned)cache_size;n++)
          if(cache_table[n].modified && cache_table[n].drive == disk)
          {
            total_count++;
            if(!count)
            {
              lba2chs(cache_table[n].sector, &disk_struct, &chs);
              if((m = abs(last_cyl - chs.cylinder)) > max_cyl_diff)
              {
                if(min_cyl_diff > m)
                  min_cyl_diff = m;
                continue;
              }
            }
            if(!count || (cache_table[n].sector == cache_table[sector_table[count-1]].sector+1
               && LONGMOD(cache_table[n].sector, disk_struct.sectors)))
            {
              sector_table[count++] = n;
            }
            else
            {
              if(LONGMOD(cache_table[n].sector, disk_struct.sectors) != disk_struct.sectors - 1
              && cache_table[n].sector == cache_table[sector_table[0]].sector - 1)
              {
                nmemcpyr(&sector_table[1], &sector_table[0], count<<1);
                count++;
                sector_table[0] = n;
              }
            }
            if(count == max_count) goto write;
          }
        if(!count)
        {
          if(index != -1)
            return;
          else
            if(total_count)
              max_cyl_diff = min_cyl_diff;
            else
              goto nextdisk;
        }
        else break;
      }
      while(1);
      write:for(n=m=0;n<count;n++,m+=sector_size)
      {
        memcpy(&raw_cache[m], MK_FP(cache_table[sector_table[n]].buff_seg, 0), sector_size);
        cache_table[sector_table[n]].modified = 0;
      }
      retry:
      retries = DISK_RETRY_COUNT;
      while(1)
      {
        if(disk_struct.sectors != 0xFF)
        {
          lba2chs(cache_table[sector_table[0]].sector, &disk_struct, &chs);
          DRIVE_LAST_CYL(disk) = chs.cylinder;
          status = write_physical_sector(phys_disk, &chs, count) >> 8;
        }
        else
          status = write_lba(cache_table[sector_table[0]].sector, phys_disk, count);
        if(!status)
          break;
        if(!retries--)
        {
          switch(disk_critical_error(disk, 1, status))
          {
            case CRITERR_RETRY: goto retry;
            case CRITERR_IGNORE: break;
            case CRITERR_FAIL: error_code = 0x0E;
          }
          return;
        }
        else if(!(retries & 0x3))
          reset_disk_pragma(phys_disk);
      }
    }
    while(1);
  }
}

static int sector_in_rawcache(lbaindex_t sector, unsigned char const disk_id)
{
  if(RAWCACHE_LAST_DISK == disk_id)
  {
#ifdef USE_32BIT
    b32_llsub(&sector, &RAWCACHE_LAST_SECTOR);
#else
    sector -= RAWCACHE_LAST_SECTOR;
#endif
    if(sector < RAWCACHE_LAST_SECTOR_COUNT)
      return sector;
  }
  return -1;
}

int read_logical_sector(lbaindex_t sector, unsigned char const disk_id, unsigned count)
{
  struct disk disk;
  struct chs chs;
  unsigned retries;
  unsigned status;
  unsigned char * const ptr = rawcache_buff;
  unsigned const sector_size = DRIVE_SECTOR_SIZE(disk_id);
  unsigned char const phys_disk = logical2physicaldrive(disk_id);
  struct cache_retstr cache_ret;
  unsigned n = 0;
  unsigned l;
  unsigned max_count;
  unsigned s;
  unsigned char __far * const buff = buffer_segment;

  retry:
  retries = DISK_RETRY_COUNT;
  do
    status = get_disk_parm(phys_disk, &disk, WRITE_FLAG);
  while(status && --retries);

  if(!retries)
  {
    critical_error:switch(disk_critical_error(disk_id, 0, status))
    {
      case CRITERR_IGNORE: status = 0;
      case CRITERR_FAIL: break;
      case CRITERR_RETRY: goto retry;
    }
    return -status;
  }

  if(cache_seg)
  {
    read_cache(sector, buff, disk_id, count, &cache_ret);
    if(cache_ret.index != -1)
    {
      l = DRIVE_MIN_RAWBUFF_SECTORS(disk_id);
      if(count < l) l = count;
      for(n++;--l;n++)
        if(!cache_ret.found[n])
        {
#ifdef USE_32BIT
          b32_lba_adds(&sector, &n);
#else
          sector += n;
#endif
          for(count=1,n++;--l;n++)
            if(cache_ret.found[n])
              break;
            else
              count++;
          break;
        }
      if(!l) return 0;
      if(sector_in_rawcache(sector, disk_id) != -1)
        return 0;
    }
    else if((s = sector_in_rawcache(sector, disk_id)) != -1)
    {
      memcpy(buff, &ptr[s*sector_size], sector_size);
      return 0;
    }
  }

  if(cache_seg && !RAWCACHE_SAVED)
    save_rawcache();

  if(mul_16to32(count, sector_size) > rawcache_size) count = rawcache_size/sector_size;

  while(1)
  {
    if(disk.sectors != 0xFF)
    {
      lba2chs(sector, &disk, &chs);
      DRIVE_LAST_CYL(disk_id) = chs.cylinder;

      if(!(phys_disk & 0x80))
      {
        max_count = disk.sectors - LONGMOD(sector, disk.sectors);
        if(count > max_count) count = max_count; 
      }
      status = read_physical_sector(phys_disk, &chs, count) >> 8;
    }
    else
    {
      status = read_lba(sector, phys_disk, count);
    }

    if(!status) break;

    invalidate_rawcache();
    if(!retries--)
      goto critical_error;
    else if(!(retries & 3))
      reset_disk_pragma(phys_disk);      
  }

  if(!n)
    memcpy(buff, ptr, sector_size);
  set_rawcache_last(sector, disk_id, count);
  DYNFLAGS |= RAWCACHE_SAVED_BIT;

  return 0;
}

/*unsigned char physical2logicaldrive(unsigned char physical)
{
  if(!(physical & 0x80))
    return physical+1;

  physical &= 0x7F;

  physical += 3;

  if(FLOPPY_COUNT > 2)
    physical += 2;

  return physical;
}*/

unsigned char logical2physicaldrive(unsigned char logical)
{
  if(logical < FLOPPY_COUNT) return logical;

  return (logical - FLOPPY_COUNT_R) | 0x80;
}
